@@ -333,7 +333,7 @@ gssdp_resource_browser_class_init (GSSDPResourceBrowserClass *klass)
          * @resource_browser: The #GSSDPResourceBrowser that received the
          * signal
          * @usn: The USN of the discovered resource
-         * @locations: A #GList of strings describing the locations of the
+         * @locations: (type GList*) (transfer none) (element-type utf8): A #GList of strings describing the locations of the
          * discovered resource.
          *
          * The ::resource-available signal is emitted whenever a new resource
@@ -416,7 +416,7 @@ gssdp_resource_browser_set_client (GSSDPResourceBrowser *resource_browser,
  * gssdp_resource_browser_get_client
  * @resource_browser: A #GSSDPResourceBrowser
  *
- * Return value: The #GSSDPClient @resource_browser is associated with.
+ * Returns: (transfer none): The #GSSDPClient @resource_browser is associated with.
  **/
 GSSDPClient *
 gssdp_resource_browser_get_client (GSSDPResourceBrowser *resource_browser)
@@ -458,10 +458,10 @@ gssdp_resource_browser_set_target (GSSDPResourceBrowser *resource_browser,
         pattern = g_strndup (target,
                              strlen (target) + strlen (version_pattern));
 
-        version = g_strrstr (pattern, ":") + 1;
+        version = g_strrstr (pattern, ":");
         if (version != NULL &&
-            g_regex_match_simple (version_pattern, version, 0, 0)) {
-                strcpy (version, version_pattern);
+            g_regex_match_simple (version_pattern, version + 1, 0, 0)) {
+                strcpy (version + 1, version_pattern);
         }
 
         error = NULL;
@@ -580,10 +580,12 @@ gssdp_resource_browser_get_active (GSSDPResourceBrowser *resource_browser)
 static gboolean
 resource_expire (gpointer user_data)
 {
+        GSSDPResourceBrowser *resource_browser;
         Resource *resource;
         char *usn;
 
         resource = user_data;
+        resource_browser = resource->resource_browser;
 
         /* Steal the USN pointer from the resource as we need it for the signal
          * emission.
@@ -593,7 +595,7 @@ resource_expire (gpointer user_data)
 
         g_hash_table_remove (resource->resource_browser->priv->resources, usn);
 
-        g_signal_emit (resource->resource_browser,
+        g_signal_emit (resource_browser,
                        signals[RESOURCE_UNAVAILABLE],
                        0,
                        usn);
