@@ -58,18 +58,24 @@ struct _GSSDPResourceGroupPrivate {
         GSource     *timeout_src;
 
         guint        last_resource_id;
+        
+        guint        message_delay;
+        GQueue      *message_queue;
+        GSource     *message_src;
 };
 
 enum {
         PROP_0,
         PROP_CLIENT,
         PROP_MAX_AGE,
-        PROP_AVAILABLE
+        PROP_AVAILABLE,
+        PROP_MESSAGE_DELAY
 };
 
 typedef struct {
         GSSDPResourceGroup *resource_group;
 
+        GRegex              *target_regex;
         char                *target;
         char                *usn;
         GList               *locations;
@@ -77,6 +83,10 @@ typedef struct {
         GList               *responses;
 
         guint                id;
+
+        guint                version;
+
+        gboolean             initial_byebye_sent;
 } Resource;
 
 typedef struct {
@@ -88,6 +98,9 @@ typedef struct {
         GSource  *timeout_src;
 } DiscoveryResponse;
 
+#define DEFAULT_MESSAGE_DELAY 120
+#define VERSION_PATTERN "[0-9]+$"
+
 /* Function prototypes */
 static void
 gssdp_resource_group_set_client (GSSDPResourceGroup *resource_group,
@@ -111,6 +124,16 @@ static gboolean
 discovery_response_timeout      (gpointer            user_data);
 static void
 discovery_response_free         (DiscoveryResponse  *response);
+static gboolean
+process_queue                   (gpointer            data);
+static char *
+get_version_for_target          (char *target);
+static GRegex *
+create_target_regex             (const char         *target,
+                                 guint              *version,
+                                 GError            **error);
+static void
+send_initial_resource_byebye    (Resource          *resource);
 
 static void
 gssdp_resource_group_init (GSSDPResourceGroup *resource_group)
@@ -121,6 +144,9 @@ gssdp_resource_group_init (GSSDPResourceGroup *resource_group)
                                          GSSDPResourceGroupPrivate);
 
         resource_group->priv->max_age = SSDP_DEFAULT_MAX_AGE;
+        resource_group->priv->message_delay = DEFAULT_MESSAGE_DELAY;
+
+        resource_group->priv->message_queue = g_queue_new ();
 }
 
 static void
@@ -149,6 +175,12 @@ gssdp_resource_group_get_property (GObject    *object,
                         (value,
                          gssdp_resource_group_get_available (resource_group));
                 break;
+        case PROP_MESSAGE_DELAY:
+                g_value_set_uint
+                        (value,
+                         gssdp_resource_group_get_message_delay 
+                                (resource_group));
+                break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
                 break;
@@ -178,6 +210,10 @@ gssdp_resource_group_set_property (GObject      *object,
                 gssdp_resource_group_set_available
                         (resource_group, g_value_get_boolean (value));
                 break;
+        case PROP_MESSAGE_DELAY:
+                gssdp_resource_group_set_message_delay
+                        (resource_group, g_value_get_uint (value));
+                break;
         default:
                 G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
                 break;
@@ -188,33 +224,56 @@ static void
 gssdp_resource_group_dispose (GObject *object)
 {
         GSSDPResourceGroup *resource_group;
+        GSSDPResourceGroupPrivate *priv;
 
         resource_group = GSSDP_RESOURCE_GROUP (object);
+        priv = resource_group->priv;
 
-        while (resource_group->priv->resources) {
-                resource_free (resource_group->priv->resources->data);
-                resource_group->priv->resources =
-                        g_list_delete_link (resource_group->priv->resources,
-                                            resource_group->priv->resources);
+        while (priv->resources) {
+                resource_free (priv->resources->data);
+                priv->resources =
+                        g_list_delete_link (priv->resources,
+                                            priv->resources);
         }
 
-        if (resource_group->priv->timeout_src) {
-                g_source_destroy (resource_group->priv->timeout_src);
-                resource_group->priv->timeout_src = NULL;
+        if (priv->message_queue) {
+                /* send messages without usual delay */
+                while (!g_queue_is_empty (priv->message_queue)) {
+                        if (priv->available)
+                                process_queue (resource_group);
+                        else
+                                g_free (g_queue_pop_head
+                                        (priv->message_queue));
+                }
+
+                g_queue_free (priv->message_queue);
+                priv->message_queue = NULL;
         }
 
-        if (resource_group->priv->client) {
+        if (priv->message_src) {
+                g_source_destroy (priv->message_src);
+                priv->message_src = NULL;
+        }
+
+        if (priv->timeout_src) {
+                g_source_destroy (priv->timeout_src);
+                priv->timeout_src = NULL;
+        }
+
+        if (priv->client) {
                 if (g_signal_handler_is_connected
-                        (resource_group->priv->client,
-                         resource_group->priv->message_received_id)) {
+                        (priv->client,
+                         priv->message_received_id)) {
                         g_signal_handler_disconnect
-                                (resource_group->priv->client,
-                                 resource_group->priv->message_received_id);
+                                (priv->client,
+                                 priv->message_received_id);
                 }
                                                    
-                g_object_unref (resource_group->priv->client);
-                resource_group->priv->client = NULL;
+                g_object_unref (priv->client);
+                priv->client = NULL;
         }
+
+        G_OBJECT_CLASS (gssdp_resource_group_parent_class)->dispose (object);
 }
 
 static void
@@ -283,6 +342,27 @@ gssdp_resource_group_class_init (GSSDPResourceGroupClass *klass)
                           G_PARAM_READWRITE |
                           G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
                           G_PARAM_STATIC_BLURB));
+
+        /**
+         * GSSDPResourceGroup:message-delay
+         *
+         * The minimum number of milliseconds between SSDP messages.
+         * The default is 120 based on DLNA specification.
+         **/
+        g_object_class_install_property
+                (object_class,
+                 PROP_MESSAGE_DELAY,
+                 g_param_spec_uint
+                         ("message-delay",
+                          "Message delay",
+                          "The minimum number of milliseconds between SSDP "
+                          "messages.",
+                          0,
+                          G_MAXUINT,
+                          DEFAULT_MESSAGE_DELAY,
+                          G_PARAM_READWRITE |
+                          G_PARAM_STATIC_NAME | G_PARAM_STATIC_NICK |
+                          G_PARAM_STATIC_BLURB));
 }
 
 /**
@@ -325,7 +405,7 @@ gssdp_resource_group_set_client (GSSDPResourceGroup *resource_group,
  * gssdp_resource_group_get_client
  * @resource_group: A #GSSDPResourceGroup
  *
- * Return value: The #GSSDPClient @resource_group is associated with.
+ * Return: (transfer none): The #GSSDPClient @resource_group is associated with.
  **/
 GSSDPClient *
 gssdp_resource_group_get_client (GSSDPResourceGroup *resource_group)
@@ -338,7 +418,7 @@ gssdp_resource_group_get_client (GSSDPResourceGroup *resource_group)
 /**
  * gssdp_resource_group_set_max_age
  * @resource_group: A #GSSDPResourceGroup
- * @mx: The number of seconds advertisements are valid
+ * @max_age: The number of seconds advertisements are valid
  *
  * Sets the number of seconds advertisements are valid to @max_age.
  **/
@@ -370,6 +450,55 @@ gssdp_resource_group_get_max_age (GSSDPResourceGroup *resource_group)
         return resource_group->priv->max_age;
 }
 
+/**
+ * gssdp_resource_group_set_message_delay
+ * @resource_group: A #GSSDPResourceGroup
+ * @message_delay: The message delay in ms.
+ *
+ * Sets the minimum time between each SSDP message.
+ **/
+void
+gssdp_resource_group_set_message_delay (GSSDPResourceGroup *resource_group,
+                                        guint               message_delay)
+{
+        g_return_if_fail (GSSDP_IS_RESOURCE_GROUP (resource_group));
+
+        if (resource_group->priv->message_delay == message_delay)
+                return;
+
+        resource_group->priv->message_delay = message_delay;
+        
+        g_object_notify (G_OBJECT (resource_group), "message-delay");
+}
+
+/**
+ * gssdp_resource_group_get_message_delay
+ * @resource_group: A #GSSDPResourceGroup
+ *
+ * Return value: the minimum time between each SSDP message in ms.
+ **/
+guint
+gssdp_resource_group_get_message_delay (GSSDPResourceGroup *resource_group)
+{
+        g_return_val_if_fail (GSSDP_IS_RESOURCE_GROUP (resource_group), 0);
+
+        return resource_group->priv->message_delay;
+}
+
+static void
+send_initial_resource_byebye (Resource *resource)
+{
+        if (!resource->initial_byebye_sent) {
+                /* Unannounce before first announce. This is
+                   done to minimize the possibility of
+                   control points thinking that this is just
+                   a reannouncement. */
+                resource_byebye (resource);
+
+                resource->initial_byebye_sent = TRUE;
+        }
+}
+
 /**
  * gssdp_resource_group_set_available
  * @resource_group: A #GSSDPResourceGroup
@@ -396,10 +525,15 @@ gssdp_resource_group_set_available (GSSDPResourceGroup *resource_group,
                 GMainContext *context;
                 int timeout;
 
-                /* We want to re-announce before actually timing out */
+                /* We want to re-announce at least 3 times before the resource
+                 * group expires to cope with the unrelialble nature of UDP.
+                 *
+                 * Read the paragraphs about 'CACHE-CONTROL' on pages 21-22 of
+                 * UPnP Device Architecture Document v1.1 for further details.
+                 * */
                 timeout = resource_group->priv->max_age;
-                if (timeout > 2)
-                        timeout = timeout / 2 - 1;
+                if (G_LIKELY (timeout > 6))
+                        timeout = (timeout / 3) - 1;
 
                 /* Add re-announcement timer */
                 resource_group->priv->timeout_src =
@@ -414,6 +548,11 @@ gssdp_resource_group_set_available (GSSDPResourceGroup *resource_group,
 
                 g_source_unref (resource_group->priv->timeout_src);
 
+                /* Make sure initial byebyes are sent grouped before initial
+                 * alives */
+                for (l = resource_group->priv->resources; l; l = l->next)
+                        send_initial_resource_byebye (l->data);
+
                 /* Announce all resources */
                 for (l = resource_group->priv->resources; l; l = l->next)
                         resource_alive (l->data);
@@ -464,6 +603,7 @@ gssdp_resource_group_add_resource (GSSDPResourceGroup *resource_group,
 {
         Resource *resource;
         GList *l;
+        GError *error;
 
         g_return_val_if_fail (GSSDP_IS_RESOURCE_GROUP (resource_group), 0);
         g_return_val_if_fail (target != NULL, 0);
@@ -477,6 +617,21 @@ gssdp_resource_group_add_resource (GSSDPResourceGroup *resource_group,
         resource->target = g_strdup (target);
         resource->usn    = g_strdup (usn);
 
+        error = NULL;
+        resource->target_regex = create_target_regex (target, &resource->version, &error);
+        if (error) {
+                g_warning ("Error compiling regular expression for '%s': %s",
+                           target,
+                           error->message);
+
+                g_error_free (error);
+                resource_free (resource);
+
+                return 0;
+        }
+
+        resource->initial_byebye_sent = FALSE;
+
         for (l = locations; l; l = l->next) {
                 resource->locations = g_list_append (resource->locations,
                                                     g_strdup (l->data));
@@ -511,32 +666,15 @@ gssdp_resource_group_add_resource_simple (GSSDPResourceGroup *resource_group,
                                           const char         *usn,
                                           const char         *location)
 {
-        Resource *resource;
+        GList *locations = NULL;
+        guint  resource_id;
 
-        g_return_val_if_fail (GSSDP_IS_RESOURCE_GROUP (resource_group), 0);
-        g_return_val_if_fail (target != NULL, 0);
-        g_return_val_if_fail (usn != NULL, 0);
-        g_return_val_if_fail (location != NULL, 0);
+        locations = g_list_append (locations, (gpointer) location);
+        resource_id = gssdp_resource_group_add_resource (resource_group, target, usn, locations);
 
-        resource = g_slice_new0 (Resource);
+        g_list_free (locations);
 
-        resource->resource_group = resource_group;
-
-        resource->target = g_strdup (target);
-        resource->usn    = g_strdup (usn);
-
-        resource->locations = g_list_append (resource->locations,
-                                             g_strdup (location));
-
-        resource_group->priv->resources =
-                g_list_prepend (resource_group->priv->resources, resource);
-
-        resource->id = ++resource_group->priv->last_resource_id;
-
-        if (resource_group->priv->available)
-                resource_alive (resource);
-
-        return resource->id;
+        return resource_id;
 }
 
 /**
@@ -573,7 +711,7 @@ gssdp_resource_group_remove_resource (GSSDPResourceGroup *resource_group,
 }
 
 /**
- * Called every max-age seconds to re-announce all resources
+ * Called to re-announce all resources periodically
  **/
 static gboolean
 resource_group_timeout (gpointer user_data)
@@ -602,9 +740,9 @@ message_received_cb (GSSDPClient        *client,
                      gpointer            user_data)
 {
         GSSDPResourceGroup *resource_group;
-        const char *target, *mx_str;
+        const char *target, *mx_str, *version_str;
         gboolean want_all;
-        int mx;
+        int mx, version;
         GList *l;
 
         resource_group = GSSDP_RESOURCE_GROUP (user_data);
@@ -618,7 +756,7 @@ message_received_cb (GSSDPClient        *client,
                 return;
 
         /* Extract target */
-        target = soup_message_headers_get (headers, "ST");
+        target = soup_message_headers_get_one (headers, "ST");
         if (!target) {
                 g_warning ("Discovery request did not have an ST header");
 
@@ -629,11 +767,21 @@ message_received_cb (GSSDPClient        *client,
         want_all = (strcmp (target, GSSDP_ALL_RESOURCES) == 0);
 
         /* Extract MX */
-        mx_str = soup_message_headers_get (headers, "MX");
-        if (mx_str)
-                mx = atoi (mx_str);
+        mx_str = soup_message_headers_get_one (headers, "MX");
+        if (!mx_str || atoi (mx_str) == 0) {
+                g_warning ("Discovery request did not have a valid MX header");
+ 
+                return;
+        }
+ 
+        mx = atoi (mx_str);
+
+        /* Extract version */
+        version_str = get_version_for_target ((char *) target);
+        if (version_str != NULL)
+                version = atoi (version_str);
         else
-                mx = SSDP_DEFAULT_MX;
+                version = 0;
 
         /* Find matching resource */
         for (l = resource_group->priv->resources; l; l = l->next) {
@@ -641,7 +789,12 @@ message_received_cb (GSSDPClient        *client,
 
                 resource = l->data;
 
-                if (want_all || strcmp (resource->target, target) == 0) {
+                if (want_all ||
+                    (g_regex_match (resource->target_regex,
+                                    target,
+                                    0,
+                                    NULL) &&
+                     version <= resource->version)) {
                         /* Match. */
                         guint timeout;
                         DiscoveryResponse *response;
@@ -655,9 +808,13 @@ message_received_cb (GSSDPClient        *client,
                         
                         response->dest_ip   = g_strdup (from_ip);
                         response->dest_port = from_port;
-                        response->target    = g_strdup (target);
                         response->resource  = resource;
 
+                        if (want_all)
+                                response->target = g_strdup (resource->target);
+                        else
+                                response->target = g_strdup (target);
+
                         /* Add timeout */
                         response->timeout_src = g_timeout_source_new (timeout);
 			g_source_set_callback (response->timeout_src,
@@ -701,6 +858,25 @@ construct_al (Resource *resource)
                 return NULL; 
 }
 
+static char *
+construct_usn (const char *usn, const char *response_target, const char *resource_target)
+{
+        const char *needle;
+        char *prefix;
+        char *st;
+
+        needle = strstr (usn, resource_target);
+        if (needle == NULL)
+                return g_strdup (usn);
+
+        prefix = g_strndup (usn, needle - usn);
+        st = g_strconcat (prefix, response_target, NULL);
+
+        g_free (prefix);
+
+        return st;
+}
+
 /**
  * Send a discovery response
  **/
@@ -712,6 +888,7 @@ discovery_response_timeout (gpointer user_data)
         SoupDate *date;
         char *al, *date_str, *message;
         guint max_age;
+        char *usn;
 
         response = user_data;
 
@@ -721,7 +898,7 @@ discovery_response_timeout (gpointer user_data)
         max_age = response->resource->resource_group->priv->max_age;
 
         al = construct_al (response->resource);
-
+        usn = construct_usn (response->resource->usn, response->target, response->resource->target);
         date = soup_date_new_from_now (0);
         date_str = soup_date_to_string (date, SOUP_DATE_HTTP);
         soup_date_free (date);
@@ -729,7 +906,7 @@ discovery_response_timeout (gpointer user_data)
         message = g_strdup_printf (SSDP_DISCOVERY_RESPONSE,
                                    (char *) response->resource->locations->data,
                                    al ? al : "",
-                                   response->resource->usn,
+                                   usn,
                                    gssdp_client_get_server_id (client),
                                    max_age,
                                    response->target,
@@ -743,6 +920,7 @@ discovery_response_timeout (gpointer user_data)
         g_free (message);
         g_free (date_str);
         g_free (al);
+        g_free (usn);
 
         discovery_response_free (response);
 
@@ -766,6 +944,68 @@ discovery_response_free (DiscoveryResponse *response)
         g_slice_free (DiscoveryResponse, response);
 }
 
+/**
+ * Send the next queued message, if any
+ **/
+static gboolean
+process_queue (gpointer data)
+{
+        GSSDPResourceGroup *resource_group;
+
+        resource_group = GSSDP_RESOURCE_GROUP (data);
+
+        if (g_queue_is_empty (resource_group->priv->message_queue)) {
+                /* this is the timeout after last message in queue */
+                resource_group->priv->message_src = NULL;
+
+                return FALSE;
+        } else {
+                GSSDPClient *client;
+                char *message;
+
+                client = resource_group->priv->client;
+                message = g_queue_pop_head
+                        (resource_group->priv->message_queue);
+
+                _gssdp_client_send_message (client,
+                                            NULL,
+                                            0,
+                                            message);
+                g_free (message);
+
+                return TRUE;
+        }
+}
+
+/**
+ * Add a message to sending queue
+ * 
+ * Do not free @message.
+ **/
+static void
+queue_message (GSSDPResourceGroup *resource_group,
+               char               *message)
+{
+        g_queue_push_tail (resource_group->priv->message_queue, 
+                           message);
+
+        if (resource_group->priv->message_src == NULL) {
+                /* nothing in the queue: process message immediately 
+                   and add a timeout for (possible) next message */
+                GMainContext *context;
+
+                process_queue (resource_group);
+                resource_group->priv->message_src = g_timeout_source_new (
+                    resource_group->priv->message_delay);
+                g_source_set_callback (resource_group->priv->message_src,
+                    process_queue, resource_group, NULL);
+                context = gssdp_client_get_main_context (
+                    resource_group->priv->client);
+                g_source_attach (resource_group->priv->message_src, context);
+                g_source_unref (resource_group->priv->message_src);
+        }
+}
+
 /**
  * Send ssdp:alive message for @resource
  **/
@@ -775,6 +1015,10 @@ resource_alive (Resource *resource)
         GSSDPClient *client;
         guint max_age;
         char *al, *message;
+        guint8 i;
+
+        /* Send initial byebye if not sent already */
+        send_initial_resource_byebye (resource);
 
         /* Send message */
         client = resource->resource_group->priv->client;
@@ -783,20 +1027,18 @@ resource_alive (Resource *resource)
 
         al = construct_al (resource);
 
-        message = g_strdup_printf (SSDP_ALIVE_MESSAGE,
-                                   max_age,
-                                   (char *) resource->locations->data,
-                                   al ? al : "",
-                                   gssdp_client_get_server_id (client),
-                                   resource->target,
-                                   resource->usn);
+        for (i = 0; i < 3; i++) {
+                message = g_strdup_printf (SSDP_ALIVE_MESSAGE,
+                                           max_age,
+                                           (char *) resource->locations->data,
+                                           al ? al : "",
+                                           gssdp_client_get_server_id (client),
+                                           resource->target,
+                                           resource->usn);
 
-        _gssdp_client_send_message (client,
-                                    NULL,
-                                    0,
-                                    message);
+                queue_message (resource->resource_group, message);
+        }
 
-        g_free (message);
         g_free (al);
 }
 
@@ -806,22 +1048,17 @@ resource_alive (Resource *resource)
 static void
 resource_byebye (Resource *resource)
 {
-        GSSDPClient *client;
+        guint8 i;
         char *message;
 
-        /* Send message */
-        client = resource->resource_group->priv->client;
+        for (i = 0; i < 3; i++) {
+                /* Queue message */
+                message = g_strdup_printf (SSDP_BYEBYE_MESSAGE,
+                                           resource->target,
+                                           resource->usn);
 
-        message = g_strdup_printf (SSDP_BYEBYE_MESSAGE,
-                                   resource->target,
-                                   resource->usn);
-        
-        _gssdp_client_send_message (client,
-                                    NULL,
-                                    0,
-                                    message);
-
-        g_free (message);
+                queue_message (resource->resource_group, message);
+        }
 }
 
 /**
@@ -839,6 +1076,9 @@ resource_free (Resource *resource)
         g_free (resource->usn);
         g_free (resource->target);
 
+        if (resource->target_regex)
+                g_regex_unref (resource->target_regex);
+
         while (resource->locations) {
                 g_free (resource->locations->data);
                 resource->locations = g_list_delete_link (resource->locations,
@@ -847,3 +1087,48 @@ resource_free (Resource *resource)
 
         g_slice_free (Resource, resource);
 }
+
+/* Gets you the pointer to the version part in the target string */
+static char *
+get_version_for_target (char *target)
+{
+        char *version;
+
+        if (strncmp (target, "urn:", 4) != 0) {
+                /* target is not a URN so no version. */
+                return NULL;
+        }
+
+        version = g_strrstr (target, ":") + 1;
+        if (version == NULL ||
+            !g_regex_match_simple (VERSION_PATTERN, version, 0, 0))
+                return NULL;
+
+        return version;
+}
+
+static GRegex *
+create_target_regex (const char *target, guint *version, GError **error)
+{
+        GRegex *regex;
+        char *pattern;
+        char *version_str;
+
+        *version = 0;
+        /* Make sure we have enough room for version pattern */
+        pattern = g_strndup (target,
+                             strlen (target) + strlen (VERSION_PATTERN));
+
+        version_str = get_version_for_target (pattern);
+        if (version_str != NULL) {
+                *version = atoi (version_str);
+                strcpy (version_str, VERSION_PATTERN);
+        }
+
+        regex = g_regex_new (pattern, 0, 0, error);
+
+        g_free (pattern);
+
+        return regex;
+}
+
